// Copyright 2014 The Flutter Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

import 'package:flutter/gestures.dart';

const List<PointerEvent> velocityEventData = <PointerEvent>[
  PointerDownEvent(
    timeStamp: Duration(milliseconds: 216690896),
    pointer: 1,
    position: Offset(270.0, 538.2857055664062),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216690906),
    pointer: 1,
    position: Offset(270.0, 538.2857055664062),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216690951),
    pointer: 1,
    position: Offset(270.0, 530.8571166992188),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216690959),
    pointer: 1,
    position: Offset(270.0, 526.8571166992188),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216690967),
    pointer: 1,
    position: Offset(270.0, 521.4285888671875),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216690975),
    pointer: 1,
    position: Offset(270.0, 515.4285888671875),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216690983),
    pointer: 1,
    position: Offset(270.0, 506.8571472167969),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216690991),
    pointer: 1,
    position: Offset(268.8571472167969, 496.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216690998),
    pointer: 1,
    position: Offset(267.4285583496094, 483.1428527832031),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691006),
    pointer: 1,
    position: Offset(266.28570556640625, 469.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691014),
    pointer: 1,
    position: Offset(265.4285583496094, 456.8571472167969),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691021),
    pointer: 1,
    position: Offset(264.28570556640625, 443.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691029),
    pointer: 1,
    position: Offset(264.0, 431.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691036),
    pointer: 1,
    position: Offset(263.4285583496094, 421.1428527832031),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691044),
    pointer: 1,
    position: Offset(263.4285583496094, 412.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691052),
    pointer: 1,
    position: Offset(263.4285583496094, 404.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691060),
    pointer: 1,
    position: Offset(263.4285583496094, 396.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691068),
    pointer: 1,
    position: Offset(264.5714416503906, 390.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691075),
    pointer: 1,
    position: Offset(265.1428527832031, 384.8571472167969),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691083),
    pointer: 1,
    position: Offset(266.0, 380.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691091),
    pointer: 1,
    position: Offset(266.5714416503906, 376.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691098),
    pointer: 1,
    position: Offset(267.1428527832031, 373.1428527832031),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691106),
    pointer: 1,
    position: Offset(267.71429443359375, 370.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691114),
    pointer: 1,
    position: Offset(268.28570556640625, 367.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691121),
    pointer: 1,
    position: Offset(268.5714416503906, 366.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691130),
    pointer: 1,
    position: Offset(268.8571472167969, 364.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691137),
    pointer: 1,
    position: Offset(269.1428527832031, 363.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691145),
    pointer: 1,
    position: Offset(269.1428527832031, 362.8571472167969),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691153),
    pointer: 1,
    position: Offset(269.4285583496094, 362.8571472167969),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691168),
    pointer: 1,
    position: Offset(268.5714416503906, 365.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691176),
    pointer: 1,
    position: Offset(267.1428527832031, 370.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691183),
    pointer: 1,
    position: Offset(265.4285583496094, 376.8571472167969),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691191),
    pointer: 1,
    position: Offset(263.1428527832031, 385.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691199),
    pointer: 1,
    position: Offset(261.4285583496094, 396.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691207),
    pointer: 1,
    position: Offset(259.71429443359375, 408.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691215),
    pointer: 1,
    position: Offset(258.28570556640625, 419.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691222),
    pointer: 1,
    position: Offset(257.4285583496094, 428.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691230),
    pointer: 1,
    position: Offset(256.28570556640625, 436.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691238),
    pointer: 1,
    position: Offset(255.7142791748047, 442.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691245),
    pointer: 1,
    position: Offset(255.14285278320312, 447.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691253),
    pointer: 1,
    position: Offset(254.85714721679688, 453.1428527832031),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691261),
    pointer: 1,
    position: Offset(254.57142639160156, 458.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691268),
    pointer: 1,
    position: Offset(254.2857208251953, 463.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691276),
    pointer: 1,
    position: Offset(254.2857208251953, 470.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691284),
    pointer: 1,
    position: Offset(254.2857208251953, 477.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691292),
    pointer: 1,
    position: Offset(255.7142791748047, 487.1428527832031),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691300),
    pointer: 1,
    position: Offset(256.8571472167969, 498.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691307),
    pointer: 1,
    position: Offset(258.28570556640625, 507.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691315),
    pointer: 1,
    position: Offset(259.4285583496094, 516.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691323),
    pointer: 1,
    position: Offset(260.28570556640625, 521.7142944335938),
  ),
  PointerUpEvent(
    timeStamp: Duration(milliseconds: 216691338),
    pointer: 1,
    position: Offset(260.28570556640625, 521.7142944335938),
  ),
  PointerDownEvent(
    timeStamp: Duration(milliseconds: 216691573),
    pointer: 2,
    position: Offset(266.0, 327.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691588),
    pointer: 2,
    position: Offset(266.0, 327.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691626),
    pointer: 2,
    position: Offset(261.1428527832031, 337.1428527832031),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691634),
    pointer: 2,
    position: Offset(258.28570556640625, 343.1428527832031),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691642),
    pointer: 2,
    position: Offset(254.57142639160156, 354.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691650),
    pointer: 2,
    position: Offset(250.2857208251953, 368.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691657),
    pointer: 2,
    position: Offset(247.42857360839844, 382.8571472167969),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691665),
    pointer: 2,
    position: Offset(245.14285278320312, 397.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691673),
    pointer: 2,
    position: Offset(243.14285278320312, 411.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691680),
    pointer: 2,
    position: Offset(242.2857208251953, 426.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691688),
    pointer: 2,
    position: Offset(241.7142791748047, 440.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691696),
    pointer: 2,
    position: Offset(241.7142791748047, 454.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691703),
    pointer: 2,
    position: Offset(242.57142639160156, 467.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691712),
    pointer: 2,
    position: Offset(243.42857360839844, 477.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691720),
    pointer: 2,
    position: Offset(244.85714721679688, 485.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691727),
    pointer: 2,
    position: Offset(246.2857208251953, 493.1428527832031),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216691735),
    pointer: 2,
    position: Offset(248.0, 499.71429443359375),
  ),
  PointerUpEvent(
    timeStamp: Duration(milliseconds: 216691750),
    pointer: 2,
    position: Offset(248.0, 499.71429443359375),
  ),
  PointerDownEvent(
    timeStamp: Duration(milliseconds: 216692255),
    pointer: 3,
    position: Offset(249.42857360839844, 351.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692270),
    pointer: 3,
    position: Offset(249.42857360839844, 351.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692309),
    pointer: 3,
    position: Offset(246.2857208251953, 361.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692317),
    pointer: 3,
    position: Offset(244.0, 368.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692325),
    pointer: 3,
    position: Offset(241.42857360839844, 377.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692333),
    pointer: 3,
    position: Offset(237.7142791748047, 391.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692340),
    pointer: 3,
    position: Offset(235.14285278320312, 406.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692348),
    pointer: 3,
    position: Offset(232.57142639160156, 421.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692356),
    pointer: 3,
    position: Offset(230.2857208251953, 436.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692363),
    pointer: 3,
    position: Offset(228.2857208251953, 451.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692371),
    pointer: 3,
    position: Offset(227.42857360839844, 466.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692378),
    pointer: 3,
    position: Offset(226.2857208251953, 479.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692387),
    pointer: 3,
    position: Offset(225.7142791748047, 491.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692395),
    pointer: 3,
    position: Offset(225.14285278320312, 501.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692402),
    pointer: 3,
    position: Offset(224.85714721679688, 509.1428527832031),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692410),
    pointer: 3,
    position: Offset(224.57142639160156, 514.8571166992188),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692418),
    pointer: 3,
    position: Offset(224.2857208251953, 519.4285888671875),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692425),
    pointer: 3,
    position: Offset(224.0, 523.4285888671875),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692433),
    pointer: 3,
    position: Offset(224.0, 527.1428833007812),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692441),
    pointer: 3,
    position: Offset(224.0, 530.5714111328125),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692448),
    pointer: 3,
    position: Offset(224.0, 533.1428833007812),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692456),
    pointer: 3,
    position: Offset(224.0, 535.4285888671875),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692464),
    pointer: 3,
    position: Offset(223.7142791748047, 536.8571166992188),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692472),
    pointer: 3,
    position: Offset(223.7142791748047, 538.2857055664062),
  ),
  PointerUpEvent(
    timeStamp: Duration(milliseconds: 216692487),
    pointer: 3,
    position: Offset(223.7142791748047, 538.2857055664062),
  ),
  PointerDownEvent(
    timeStamp: Duration(milliseconds: 216692678),
    pointer: 4,
    position: Offset(221.42857360839844, 526.2857055664062),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692701),
    pointer: 4,
    position: Offset(220.57142639160156, 514.8571166992188),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692708),
    pointer: 4,
    position: Offset(220.2857208251953, 508.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692716),
    pointer: 4,
    position: Offset(220.2857208251953, 498.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692724),
    pointer: 4,
    position: Offset(221.14285278320312, 484.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692732),
    pointer: 4,
    position: Offset(221.7142791748047, 469.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692740),
    pointer: 4,
    position: Offset(223.42857360839844, 453.1428527832031),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692748),
    pointer: 4,
    position: Offset(225.7142791748047, 436.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692755),
    pointer: 4,
    position: Offset(229.14285278320312, 418.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692763),
    pointer: 4,
    position: Offset(232.85714721679688, 400.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692770),
    pointer: 4,
    position: Offset(236.85714721679688, 382.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692778),
    pointer: 4,
    position: Offset(241.14285278320312, 366.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692786),
    pointer: 4,
    position: Offset(244.85714721679688, 350.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216692793),
    pointer: 4,
    position: Offset(249.14285278320312, 335.4285583496094),
  ),
  PointerUpEvent(
    timeStamp: Duration(milliseconds: 216692809),
    pointer: 4,
    position: Offset(249.14285278320312, 335.4285583496094),
  ),
  PointerDownEvent(
    timeStamp: Duration(milliseconds: 216693222),
    pointer: 5,
    position: Offset(224.0, 545.4285888671875),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216693245),
    pointer: 5,
    position: Offset(224.0, 545.4285888671875),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216693275),
    pointer: 5,
    position: Offset(222.85714721679688, 535.1428833007812),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216693284),
    pointer: 5,
    position: Offset(222.85714721679688, 528.8571166992188),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216693291),
    pointer: 5,
    position: Offset(222.2857208251953, 518.5714111328125),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216693299),
    pointer: 5,
    position: Offset(222.0, 503.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216693307),
    pointer: 5,
    position: Offset(222.0, 485.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216693314),
    pointer: 5,
    position: Offset(221.7142791748047, 464.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216693322),
    pointer: 5,
    position: Offset(222.2857208251953, 440.28570556640625),
  ),
  PointerUpEvent(
    timeStamp: Duration(milliseconds: 216693337),
    pointer: 5,
    position: Offset(222.2857208251953, 440.28570556640625),
  ),
  PointerDownEvent(
    timeStamp: Duration(milliseconds: 216693985),
    pointer: 6,
    position: Offset(208.0, 544.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694047),
    pointer: 6,
    position: Offset(208.57142639160156, 532.2857055664062),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694054),
    pointer: 6,
    position: Offset(208.85714721679688, 525.7142944335938),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694062),
    pointer: 6,
    position: Offset(208.85714721679688, 515.1428833007812),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694070),
    pointer: 6,
    position: Offset(208.0, 501.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694077),
    pointer: 6,
    position: Offset(207.42857360839844, 487.1428527832031),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694085),
    pointer: 6,
    position: Offset(206.57142639160156, 472.8571472167969),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694092),
    pointer: 6,
    position: Offset(206.57142639160156, 458.8571472167969),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694100),
    pointer: 6,
    position: Offset(206.57142639160156, 446.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694108),
    pointer: 6,
    position: Offset(206.57142639160156, 434.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694116),
    pointer: 6,
    position: Offset(207.14285278320312, 423.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694124),
    pointer: 6,
    position: Offset(208.57142639160156, 412.8571472167969),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694131),
    pointer: 6,
    position: Offset(209.7142791748047, 402.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694139),
    pointer: 6,
    position: Offset(211.7142791748047, 393.1428527832031),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694147),
    pointer: 6,
    position: Offset(213.42857360839844, 385.1428527832031),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694154),
    pointer: 6,
    position: Offset(215.42857360839844, 378.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694162),
    pointer: 6,
    position: Offset(217.42857360839844, 371.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694169),
    pointer: 6,
    position: Offset(219.42857360839844, 366.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694177),
    pointer: 6,
    position: Offset(221.42857360839844, 360.8571472167969),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694185),
    pointer: 6,
    position: Offset(223.42857360839844, 356.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694193),
    pointer: 6,
    position: Offset(225.14285278320312, 352.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694201),
    pointer: 6,
    position: Offset(226.85714721679688, 348.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694209),
    pointer: 6,
    position: Offset(228.2857208251953, 346.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694216),
    pointer: 6,
    position: Offset(229.14285278320312, 343.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694224),
    pointer: 6,
    position: Offset(230.0, 342.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694232),
    pointer: 6,
    position: Offset(230.57142639160156, 340.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694239),
    pointer: 6,
    position: Offset(230.85714721679688, 339.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694247),
    pointer: 6,
    position: Offset(230.85714721679688, 339.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694262),
    pointer: 6,
    position: Offset(230.2857208251953, 342.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694270),
    pointer: 6,
    position: Offset(228.85714721679688, 346.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694278),
    pointer: 6,
    position: Offset(227.14285278320312, 352.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694286),
    pointer: 6,
    position: Offset(225.42857360839844, 359.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694294),
    pointer: 6,
    position: Offset(223.7142791748047, 367.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694301),
    pointer: 6,
    position: Offset(222.57142639160156, 376.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694309),
    pointer: 6,
    position: Offset(221.42857360839844, 384.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694317),
    pointer: 6,
    position: Offset(220.85714721679688, 392.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694324),
    pointer: 6,
    position: Offset(220.0, 400.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694332),
    pointer: 6,
    position: Offset(219.14285278320312, 409.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694339),
    pointer: 6,
    position: Offset(218.85714721679688, 419.1428527832031),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694348),
    pointer: 6,
    position: Offset(218.2857208251953, 428.8571472167969),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694356),
    pointer: 6,
    position: Offset(218.2857208251953, 438.8571472167969),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694363),
    pointer: 6,
    position: Offset(218.2857208251953, 447.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694371),
    pointer: 6,
    position: Offset(218.2857208251953, 455.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694379),
    pointer: 6,
    position: Offset(219.14285278320312, 462.8571472167969),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694386),
    pointer: 6,
    position: Offset(220.0, 469.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694394),
    pointer: 6,
    position: Offset(221.14285278320312, 475.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694401),
    pointer: 6,
    position: Offset(222.0, 480.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694409),
    pointer: 6,
    position: Offset(222.85714721679688, 485.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694417),
    pointer: 6,
    position: Offset(224.0, 489.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694425),
    pointer: 6,
    position: Offset(224.85714721679688, 492.8571472167969),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694433),
    pointer: 6,
    position: Offset(225.42857360839844, 495.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694440),
    pointer: 6,
    position: Offset(226.0, 497.1428527832031),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694448),
    pointer: 6,
    position: Offset(226.2857208251953, 498.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694456),
    pointer: 6,
    position: Offset(226.2857208251953, 498.8571472167969),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694471),
    pointer: 6,
    position: Offset(226.2857208251953, 498.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694479),
    pointer: 6,
    position: Offset(226.2857208251953, 496.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694486),
    pointer: 6,
    position: Offset(226.2857208251953, 493.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694494),
    pointer: 6,
    position: Offset(226.2857208251953, 490.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694502),
    pointer: 6,
    position: Offset(226.2857208251953, 486.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694510),
    pointer: 6,
    position: Offset(226.2857208251953, 480.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694518),
    pointer: 6,
    position: Offset(226.2857208251953, 475.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694525),
    pointer: 6,
    position: Offset(226.2857208251953, 468.8571472167969),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694533),
    pointer: 6,
    position: Offset(226.2857208251953, 461.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694541),
    pointer: 6,
    position: Offset(226.2857208251953, 452.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694548),
    pointer: 6,
    position: Offset(226.57142639160156, 442.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694556),
    pointer: 6,
    position: Offset(226.57142639160156, 432.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694564),
    pointer: 6,
    position: Offset(226.85714721679688, 423.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694571),
    pointer: 6,
    position: Offset(227.42857360839844, 416.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694580),
    pointer: 6,
    position: Offset(227.7142791748047, 410.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694587),
    pointer: 6,
    position: Offset(228.2857208251953, 404.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694595),
    pointer: 6,
    position: Offset(228.85714721679688, 399.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694603),
    pointer: 6,
    position: Offset(229.14285278320312, 395.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694610),
    pointer: 6,
    position: Offset(229.42857360839844, 392.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694618),
    pointer: 6,
    position: Offset(229.7142791748047, 390.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694625),
    pointer: 6,
    position: Offset(229.7142791748047, 388.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694633),
    pointer: 6,
    position: Offset(229.7142791748047, 386.8571472167969),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694641),
    pointer: 6,
    position: Offset(229.7142791748047, 386.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694648),
    pointer: 6,
    position: Offset(229.7142791748047, 386.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694657),
    pointer: 6,
    position: Offset(228.85714721679688, 386.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694665),
    pointer: 6,
    position: Offset(228.0, 388.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694672),
    pointer: 6,
    position: Offset(226.0, 392.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694680),
    pointer: 6,
    position: Offset(224.0, 397.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694688),
    pointer: 6,
    position: Offset(222.0, 404.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694695),
    pointer: 6,
    position: Offset(219.7142791748047, 411.1428527832031),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694703),
    pointer: 6,
    position: Offset(218.2857208251953, 418.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694710),
    pointer: 6,
    position: Offset(217.14285278320312, 425.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694718),
    pointer: 6,
    position: Offset(215.7142791748047, 433.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694726),
    pointer: 6,
    position: Offset(214.85714721679688, 442.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694734),
    pointer: 6,
    position: Offset(214.0, 454.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694742),
    pointer: 6,
    position: Offset(214.0, 469.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694749),
    pointer: 6,
    position: Offset(215.42857360839844, 485.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694757),
    pointer: 6,
    position: Offset(217.7142791748047, 502.8571472167969),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694765),
    pointer: 6,
    position: Offset(221.14285278320312, 521.4285888671875),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694772),
    pointer: 6,
    position: Offset(224.57142639160156, 541.1428833007812),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694780),
    pointer: 6,
    position: Offset(229.14285278320312, 561.1428833007812),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216694788),
    pointer: 6,
    position: Offset(233.42857360839844, 578.8571166992188),
  ),
  PointerUpEvent(
    timeStamp: Duration(milliseconds: 216694802),
    pointer: 6,
    position: Offset(233.42857360839844, 578.8571166992188),
  ),
  PointerDownEvent(
    timeStamp: Duration(milliseconds: 216695344),
    pointer: 7,
    position: Offset(253.42857360839844, 310.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216695352),
    pointer: 7,
    position: Offset(253.42857360839844, 310.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216695359),
    pointer: 7,
    position: Offset(252.85714721679688, 318.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216695367),
    pointer: 7,
    position: Offset(251.14285278320312, 322.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216695375),
    pointer: 7,
    position: Offset(248.85714721679688, 327.1428527832031),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216695382),
    pointer: 7,
    position: Offset(246.0, 334.8571472167969),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216695390),
    pointer: 7,
    position: Offset(242.57142639160156, 344.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216695397),
    pointer: 7,
    position: Offset(238.85714721679688, 357.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216695406),
    pointer: 7,
    position: Offset(235.7142791748047, 371.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216695414),
    pointer: 7,
    position: Offset(232.2857208251953, 386.8571472167969),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216695421),
    pointer: 7,
    position: Offset(229.42857360839844, 402.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216695429),
    pointer: 7,
    position: Offset(227.42857360839844, 416.8571472167969),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216695437),
    pointer: 7,
    position: Offset(226.2857208251953, 431.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216695444),
    pointer: 7,
    position: Offset(226.2857208251953, 446.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216695452),
    pointer: 7,
    position: Offset(227.7142791748047, 460.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216695459),
    pointer: 7,
    position: Offset(230.0, 475.1428527832031),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216695467),
    pointer: 7,
    position: Offset(232.2857208251953, 489.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216695475),
    pointer: 7,
    position: Offset(235.7142791748047, 504.0),
  ),
  PointerUpEvent(
    timeStamp: Duration(milliseconds: 216695490),
    pointer: 7,
    position: Offset(235.7142791748047, 504.0),
  ),
  PointerDownEvent(
    timeStamp: Duration(milliseconds: 216695885),
    pointer: 8,
    position: Offset(238.85714721679688, 524.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216695908),
    pointer: 8,
    position: Offset(236.2857208251953, 515.7142944335938),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216695916),
    pointer: 8,
    position: Offset(234.85714721679688, 509.1428527832031),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216695924),
    pointer: 8,
    position: Offset(232.57142639160156, 498.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216695931),
    pointer: 8,
    position: Offset(230.57142639160156, 483.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216695939),
    pointer: 8,
    position: Offset(229.14285278320312, 466.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216695947),
    pointer: 8,
    position: Offset(229.14285278320312, 446.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216695955),
    pointer: 8,
    position: Offset(230.57142639160156, 424.8571472167969),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216695963),
    pointer: 8,
    position: Offset(232.57142639160156, 402.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216695970),
    pointer: 8,
    position: Offset(235.14285278320312, 380.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216695978),
    pointer: 8,
    position: Offset(238.57142639160156, 359.4285583496094),
  ),
  PointerUpEvent(
    timeStamp: Duration(milliseconds: 216695993),
    pointer: 8,
    position: Offset(238.57142639160156, 359.4285583496094),
  ),
  PointerDownEvent(
    timeStamp: Duration(milliseconds: 216696429),
    pointer: 9,
    position: Offset(238.2857208251953, 568.5714111328125),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216696459),
    pointer: 9,
    position: Offset(234.0, 560.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216696467),
    pointer: 9,
    position: Offset(231.42857360839844, 553.1428833007812),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216696475),
    pointer: 9,
    position: Offset(228.2857208251953, 543.1428833007812),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216696483),
    pointer: 9,
    position: Offset(225.42857360839844, 528.8571166992188),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216696491),
    pointer: 9,
    position: Offset(223.14285278320312, 512.2857055664062),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216696498),
    pointer: 9,
    position: Offset(222.0, 495.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216696506),
    pointer: 9,
    position: Offset(221.7142791748047, 477.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216696514),
    pointer: 9,
    position: Offset(221.7142791748047, 458.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216696521),
    pointer: 9,
    position: Offset(223.14285278320312, 438.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216696529),
    pointer: 9,
    position: Offset(224.2857208251953, 416.28570556640625),
  ),
  PointerUpEvent(
    timeStamp: Duration(milliseconds: 216696544),
    pointer: 9,
    position: Offset(224.2857208251953, 416.28570556640625),
  ),
  PointerDownEvent(
    timeStamp: Duration(milliseconds: 216696974),
    pointer: 10,
    position: Offset(218.57142639160156, 530.5714111328125),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216697012),
    pointer: 10,
    position: Offset(220.2857208251953, 522.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216697020),
    pointer: 10,
    position: Offset(221.14285278320312, 517.7142944335938),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216697028),
    pointer: 10,
    position: Offset(222.2857208251953, 511.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216697036),
    pointer: 10,
    position: Offset(224.0, 504.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216697044),
    pointer: 10,
    position: Offset(227.14285278320312, 490.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216697052),
    pointer: 10,
    position: Offset(229.42857360839844, 474.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216697059),
    pointer: 10,
    position: Offset(231.42857360839844, 454.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216697067),
    pointer: 10,
    position: Offset(233.7142791748047, 431.1428527832031),
  ),
  PointerUpEvent(
    timeStamp: Duration(milliseconds: 216697082),
    pointer: 10,
    position: Offset(233.7142791748047, 431.1428527832031),
  ),
  PointerDownEvent(
    timeStamp: Duration(milliseconds: 216697435),
    pointer: 11,
    position: Offset(257.1428527832031, 285.1428527832031),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216697465),
    pointer: 11,
    position: Offset(251.7142791748047, 296.8571472167969),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216697473),
    pointer: 11,
    position: Offset(248.2857208251953, 304.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216697481),
    pointer: 11,
    position: Offset(244.57142639160156, 314.8571472167969),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216697489),
    pointer: 11,
    position: Offset(240.2857208251953, 329.1428527832031),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216697497),
    pointer: 11,
    position: Offset(236.85714721679688, 345.1428527832031),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216697505),
    pointer: 11,
    position: Offset(233.7142791748047, 361.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216697512),
    pointer: 11,
    position: Offset(231.14285278320312, 378.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216697520),
    pointer: 11,
    position: Offset(229.42857360839844, 395.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216697528),
    pointer: 11,
    position: Offset(229.42857360839844, 412.8571472167969),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216697535),
    pointer: 11,
    position: Offset(230.85714721679688, 430.8571472167969),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216697543),
    pointer: 11,
    position: Offset(233.42857360839844, 449.71429443359375),
  ),
  PointerUpEvent(
    timeStamp: Duration(milliseconds: 216697558),
    pointer: 11,
    position: Offset(233.42857360839844, 449.71429443359375),
  ),
  PointerDownEvent(
    timeStamp: Duration(milliseconds: 216697749),
    pointer: 12,
    position: Offset(246.0, 311.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216697780),
    pointer: 12,
    position: Offset(244.57142639160156, 318.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216697787),
    pointer: 12,
    position: Offset(243.14285278320312, 325.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216697795),
    pointer: 12,
    position: Offset(241.42857360839844, 336.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216697803),
    pointer: 12,
    position: Offset(239.7142791748047, 351.1428527832031),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216697811),
    pointer: 12,
    position: Offset(238.2857208251953, 368.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216697819),
    pointer: 12,
    position: Offset(238.0, 389.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216697826),
    pointer: 12,
    position: Offset(239.14285278320312, 412.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216697834),
    pointer: 12,
    position: Offset(242.2857208251953, 438.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216697842),
    pointer: 12,
    position: Offset(247.42857360839844, 466.8571472167969),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216697849),
    pointer: 12,
    position: Offset(254.2857208251953, 497.71429443359375),
  ),
  PointerUpEvent(
    timeStamp: Duration(milliseconds: 216697864),
    pointer: 12,
    position: Offset(254.2857208251953, 497.71429443359375),
  ),
  PointerDownEvent(
    timeStamp: Duration(milliseconds: 216698321),
    pointer: 13,
    position: Offset(250.0, 306.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216698328),
    pointer: 13,
    position: Offset(250.0, 306.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216698344),
    pointer: 13,
    position: Offset(249.14285278320312, 314.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216698351),
    pointer: 13,
    position: Offset(247.42857360839844, 319.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216698359),
    pointer: 13,
    position: Offset(245.14285278320312, 326.8571472167969),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216698366),
    pointer: 13,
    position: Offset(241.7142791748047, 339.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216698374),
    pointer: 13,
    position: Offset(238.57142639160156, 355.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216698382),
    pointer: 13,
    position: Offset(236.2857208251953, 374.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216698390),
    pointer: 13,
    position: Offset(235.14285278320312, 396.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216698398),
    pointer: 13,
    position: Offset(236.57142639160156, 421.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216698406),
    pointer: 13,
    position: Offset(241.14285278320312, 451.4285583496094),
  ),
  PointerUpEvent(
    timeStamp: Duration(milliseconds: 216698421),
    pointer: 13,
    position: Offset(241.14285278320312, 451.4285583496094),
  ),
];

const List<PointerEvent> interruptedVelocityEventData = <PointerEvent>[
  PointerDownEvent(
    timeStamp: Duration(milliseconds: 216698321),
    pointer: 13,
    position: Offset(250.0, 306.0),
  ),

  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216698328),
    pointer: 13,
    position: Offset(250.0, 306.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216698344),
    pointer: 13,
    position: Offset(249.14285278320312, 314.0),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216698351),
    pointer: 13,
    position: Offset(247.42857360839844, 319.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216698359),
    pointer: 13,
    position: Offset(245.14285278320312, 326.8571472167969),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216698366),
    pointer: 13,
    position: Offset(241.7142791748047, 339.4285583496094),
  ),

  // The pointer "stops" here because we've introduced a 40+ms gap
  // in the move event stream. See kAssumePointerMoveStoppedMilliseconds
  // in velocity_tracker.dart.

  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216698374 + 40),
    pointer: 13,
    position: Offset(238.57142639160156, 355.71429443359375),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216698382 + 40),
    pointer: 13,
    position: Offset(236.2857208251953, 374.28570556640625),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216698390 + 40),
    pointer: 13,
    position: Offset(235.14285278320312, 396.5714416503906),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216698398 + 40),
    pointer: 13,
    position: Offset(236.57142639160156, 421.4285583496094),
  ),
  PointerMoveEvent(
    timeStamp: Duration(milliseconds: 216698406 + 40),
    pointer: 13,
    position: Offset(241.14285278320312, 451.4285583496094),
  ),
  PointerUpEvent(
    timeStamp: Duration(milliseconds: 216698421 + 40),
    pointer: 13,
    position: Offset(241.14285278320312, 451.4285583496094),
  ),
];
