// Copyright 2014 The Flutter Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

// @dart = 2.8

import 'package:flutter_driver/driver_extension.dart';

void main() {
  // Dummy. Only needed because driver needs an entry point.
  enableFlutterDriverExtension();
}
