// Copyright 2014 The Flutter Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

// @dart = 2.8

import '../test/common.dart';

void main() {
  // Tooling requires a single trivial test
  test('it runs a test', () {
    expect(true, true);
  });
}
