// Copyright 2014 The Flutter Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

import '../config_test_utils.dart';

void main() {
  testConfig('pubspec.yaml causes config scanning to stop', null);
}
