// Copyright 2014 The Flutter Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

// @dart = 2.8
import 'package:flutter_test/flutter_test.dart';

void main() {
  testWidgets('trivial', (WidgetTester tester) async {
    expect(true, true);
  });
}
