// Copyright 2014 The Flutter Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

// @dart = 2.9

import 'transitions_perf_test.dart' as transitions_perf_test;

void main([List<String> args = const <String>[]]) => transitions_perf_test.main(
  <String>['--hybrid', ...args],
);
