// Copyright 2014 The Flutter Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

/// [Flutter platform integration APIs for iOS.](https://api.flutter.dev/objcdoc/)
library iOS;
