// Copyright 2014 The Flutter Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

/// This library provides constants and matchers for testing the Android
/// accessibility implementation in a flutter driver environment.
library android_semantics_testing;

export 'src/common.dart';
export 'src/constants.dart';
export 'src/matcher.dart';
