// Copyright 2014 The Flutter Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

import 'my_app.dart' if (dart.library.html) 'my_web_app.dart';

// ignore_for_file: public_member_api_docs

void main() => startApp();
