// Copyright 2014 The Flutter Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

// @dart = 2.12

String? x;

void main() {
  print('--- TEST SUCCEEDED ---');
}
