// Copyright 2014 The Flutter Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

import '../flutter_test_alternative.dart';

void main() {
  test('should pass', () {
    expect(1 + 1, equals(2));
  });
}
